#include "mesh_loader.h"
#include "custom_hash.h"

const std::vector<mesh::node> &mesh::mesh_loader::get_node() const {
    return nodes;
}

const std::vector<mesh::finite_type> &mesh::mesh_loader::get_finite_elems() const {
    return finite_elems;
}

const std::vector<mesh::boundary_finite_element> &mesh::mesh_loader::get_boundary() const {
    return boundary;
}

std::set<int> mesh::mesh_loader::get_nodes_by_surface(int p_surface_id) const {
    std::set<int> res;


    auto first_to_find = boundary.begin();
    auto last_to_find = boundary.end();

    while (first_to_find != last_to_find) {
        auto found_boundary = std::find_if(first_to_find, last_to_find,
                                           [p_surface_id](const boundary_finite_element &cur_boundary) {
                                               return cur_boundary.surface_id == p_surface_id;
                                           });

        first_to_find = found_boundary;
        if (found_boundary != last_to_find) {
            for (const auto &node_in_boundary: found_boundary->node_id)
                res.insert(node_in_boundary);
            ++first_to_find;
        }
    }
    return res;
}

std::vector<int> mesh::mesh_loader::get_elems_by_ID(int p_node_id1, int p_node_id2, int p_node_id3) const {
    std::vector<int> res;

    auto first_to_find = finite_elems.begin();
    auto last_to_find = finite_elems.end();

    while (first_to_find != last_to_find) {
        auto found_elem = std::find_if(first_to_find, last_to_find,
                                       [p_node_id1, p_node_id2, p_node_id3](const finite_type &elem) {
                                           return std::find(elem.node_id.begin(), elem.node_id.end(), p_node_id1) !=
                                                  elem.node_id.end()
                                                  &&
                                                  std::find(elem.node_id.begin(), elem.node_id.end(), p_node_id2) !=
                                                  elem.node_id.end()
                                                  &&
                                                  std::find(elem.node_id.begin(), elem.node_id.end(), p_node_id3) !=
                                                  elem.node_id.end();
                                       });

        first_to_find = found_elem;
        if (found_elem != last_to_find) {
            res.push_back(found_elem->element_id);
            ++first_to_find;
        }
    }

    return res;
}

std::vector<int> mesh::mesh_loader::get_elems_by_edge(int p_node_id1, int p_node_id2) const {
    std::vector<int> res;

    auto first_to_find = finite_elems.begin();
    auto last_to_find = finite_elems.end();

    while (first_to_find != last_to_find) {
        auto found_elem = std::find_if(first_to_find, last_to_find,
                                       [p_node_id1, p_node_id2](const finite_type &elem) {
                                           return std::find(elem.node_id.begin(), elem.node_id.end(), p_node_id1) !=
                                                  elem.node_id.end()
                                                  &&
                                                  std::find(elem.node_id.begin(), elem.node_id.end(), p_node_id2) !=
                                                  elem.node_id.end();
                                       });

        first_to_find = found_elem;
        if (found_elem != last_to_find) {
            res.push_back(found_elem->element_id);
            ++first_to_find;
        }
    }

    return res;
}

std::vector<int> mesh::mesh_loader::get_elems_by_material(int p_material_id) const {
    std::vector<int> res;

    auto first_to_find = finite_elems.begin();
    auto last_to_find = finite_elems.end();

    while (first_to_find != last_to_find) {
        auto found_elem = std::find_if(first_to_find, last_to_find,
                                       [p_material_id](const finite_type &elem) {
                                           return elem.material_id == p_material_id;
                                       });

        first_to_find = found_elem;
        if (found_elem != last_to_find) {
            res.push_back(found_elem->element_id);
            ++first_to_find;
        }
    }

    return res;
}

std::vector<int> mesh::mesh_loader::get_boundary_by_surface(int p_surface_id) const {
    std::vector<int> res;

    auto first_to_find = boundary.begin();
    auto last_to_find = boundary.end();

    while (first_to_find != last_to_find) {
        auto found_elem = std::find_if(first_to_find, last_to_find,
                                       [p_surface_id](const boundary_finite_element &elem) {
                                           return elem.surface_id == p_surface_id;
                                       });

        first_to_find = found_elem;
        if (found_elem != last_to_find) {
            res.push_back(found_elem->element_id);
            ++first_to_find;
        }
    }

    return res;
}

void mesh::mesh_loader::insert_new_nodes() {

    std::unordered_set<edge> examined_edges;
    for (auto &elem: finite_elems) {
        std::vector<int> cur_nodes_id = elem.node_id;
        for (auto first_node = 0; first_node < 4; ++first_node)
            for (auto second_node = first_node + 1; second_node < 4; ++second_node) {
                edge cur_edge(cur_nodes_id[first_node], cur_nodes_id[second_node]);

                if (examined_edges.insert(cur_edge).second) {
                    node new_node = get_middle_node(cur_edge);
                    cur_edge.update_mid(new_node.node_id);
                    nodes.push_back(new_node);
                } else {
                    elem.node_id.push_back(cur_edge.middle_node);
                }
            }
    }

    for (auto &elem: boundary) {
        std::vector<int> cur_nodes_id = elem.node_id;
        for (auto first_node = 0; first_node < 3; ++first_node)
            for (auto second_node = first_node + 1; second_node < 3; ++second_node) {
                auto found_edge = examined_edges.find({cur_nodes_id[first_node], cur_nodes_id[second_node]});
                if (found_edge == examined_edges.end())
                    found_edge = examined_edges.find({cur_nodes_id[second_node], cur_nodes_id[first_node]});
                elem.node_id.push_back(found_edge->middle_node);
            }
    }
}

std::vector<std::set<int>> mesh::mesh_loader::get_adjacent_nodes() const {
    std::vector<std::set<int>> res(nodes.size() + 1);

    for (const auto &cur_elem: finite_elems) {
        for (auto cur_node_id: cur_elem.node_id)
            for (auto adj_node_id: cur_elem.node_id)
                if (cur_node_id != adj_node_id)
                    res[cur_node_id].insert(adj_node_id);
    }

    return res;
}

std::array<double, 3> mesh::mesh_loader::get_middle_of_edge(const edge &p_edge) {
    return
            {(get<0>(nodes[p_edge.nodes.first - 1].coors) + get<0>(nodes[p_edge.nodes.second - 1].coors)) * 1 /
             2,
             (get<1>(nodes[p_edge.nodes.first - 1].coors) + get<1>(nodes[p_edge.nodes.second - 1].coors)) * 1 /
             2,
             (get<2>(nodes[p_edge.nodes.first - 1].coors) + get<2>(nodes[p_edge.nodes.second - 1].coors)) * 1 /
             2};
}

mesh::node mesh::mesh_loader::get_middle_node(const edge &cur_edge) {
    node new_node(nodes.size() + 1, get_middle_of_edge(cur_edge), false);

    return new_node;
}

mesh::node &mesh::mesh_loader::get_node_by_ID(int p_id) {
    return nodes.at(p_id);
}

std::array<double, 3> mesh::mesh_loader::get_coors_node_by_ID(int p_id) {
    return nodes.at(p_id).coors;
}


mesh::finite_type mesh::mesh_loader::get_elems_by_min_ratio() {
    auto res = ranges::max(finite_elems,
                           std::less<>{},
                           [this](const finite_type &p_elem) {
                               return get_min_height_in_elem(p_elem) / get_max_base_in_elem(p_elem);
                           });
    return res;
}

template<typename T>
double get_area(const T &p_range) {
    auto x = p_range.at(0);
    auto y = p_range.at(1);
    auto z = p_range.at(2);

    double a = sqrt(pow(x[1] - y[1], 2) + pow(x[2] - y[2], 2) + pow(x[0] - y[0], 2));
    double b = sqrt(pow(y[1] - z[1], 2) + pow(y[2] - z[2], 2) + pow(y[0] - z[0], 2));
    double c = sqrt(pow(z[1] - x[1], 2) + pow(z[2] - x[2], 2) + pow(z[0] - x[0], 2));

    double p = (a + b + c) / 2;
    return sqrt(p * (p - a) * (p - b) * (p - c));
}

double mesh::mesh_loader::get_max_base_in_elem(const mesh::finite_type &p_elem) const {
    double max_base = 0;
    for (auto cur_node: p_elem.node_id) {
        auto _nodes_in_base = p_elem.node_id
                              | rv::filter([cur_node](auto cur_idx) { return cur_idx != cur_node; })
                              | rv::transform([this](int _idx) { return nodes[_idx - 1].coors; });
        std::vector<std::array<double, 3>> nodes_in_base;
        ranges::copy(_nodes_in_base, std::back_inserter(nodes_in_base));
        auto cur_base = get_area(nodes_in_base);
        if (cur_base > max_base)
            max_base = cur_base;
    }
    return max_base;
}

template<typename T>
double get_height(const std::array<double, 3> &p, const T &p_range) {
    auto x = p_range.at(0);
    auto y = p_range.at(1);
    auto z = p_range.at(2);

    double A = (x[1] - y[1]) * (x[2] - z[2]) - (x[1] - z[1]) * (x[2] - y[2]);
    double B = -((x[0] - y[0]) * (x[2] - z[2]) - (x[0] - z[0]) * (x[2] - y[2]));
    double C = (x[0] - y[0]) * (x[1] - z[1]) - (x[0] - z[0]) * (x[1] - y[1]);
    double D = -x[0] * A + x[1] * B - x[2] * C;

    return fabs((A * (p[0] - x[0]) + B * (p[1] - x[1]) + C * (p[2] - x[2]) + D) / sqrt(A * A + B * B + C * C));
}

double mesh::mesh_loader::get_min_height_in_elem(const finite_type &p_elem) const {

    double max_height = 0;
    for (auto cur_node: p_elem.node_id) {
        auto _nodes_in_base = p_elem.node_id
                              | rv::filter([cur_node](auto cur_idx) { return cur_idx != cur_node; })
                              | rv::transform([this](int _idx) { return nodes[_idx - 1].coors; });
        std::vector<std::array<double, 3>> nodes_in_base;
        ranges::copy(_nodes_in_base, std::back_inserter(nodes_in_base));
        double cur_height = get_height(nodes[cur_node - 1].coors, nodes_in_base);
        if (max_height < cur_height)
            max_height = cur_height;
    }

    return max_height;
}



